@@ -40,7 +40,7 @@ class DownloadJob:
     Keeps all the important data about a pending download job.
     Used for downloading whole projects but also single files.
     """
-    
+
     def __init__(self, project_path, total_size, version, update_tasks, download_queue_items, directory, mp, project_info):
         self.project_path = project_path
         self.total_size = total_size      # size of data to download (in bytes)
@@ -70,13 +70,13 @@ def _download_items(file, directory, diff_only=False):
     basename = os.path.basename(file['diff']['path']) if diff_only else os.path.basename(file['path'])
     file_size = file['diff']['size'] if diff_only else file['size']
     chunks = math.ceil(file_size / CHUNK_SIZE)
-    
+
     items = []
     for part_index in range(chunks):
         download_file_path = os.path.join(file_dir, basename + ".{}".format(part_index))
         size = min(CHUNK_SIZE, file_size - part_index * CHUNK_SIZE)
         items.append(DownloadQueueItem(file['path'], size, file['version'], diff_only, part_index, download_file_path))
-        
+
     return items
 
 
@@ -84,7 +84,7 @@ def _do_download(item, mc, mp, project_path, job):
     """ runs in worker thread """
     if job.is_cancelled:
         return
-    
+
     # TODO: make download_blocking / save_to_file cancellable so that we can cancel as soon as possible
 
     item.download_blocking(mc, mp, project_path)
@@ -121,7 +121,13 @@ def download_project_async(mc, project_path, directory, project_version=None):
     mp.log.info(f"--- start download {project_path}")
 
     try:
-        project_info = mc.project_info(project_path, version=project_version)
+        # check whether we download the latest version or not
+        latest_proj_info = mc.project_info(project_path)
+        if project_version:
+            project_info = mc.project_info(project_path, version=project_version)
+        else:
+            project_info = latest_proj_info
+
     except ClientError:
         _cleanup_failed_download(directory, mp)
         raise
@@ -135,7 +141,8 @@ def download_project_async(mc, project_path, directory, project_version=None):
     for file in project_info['files']:
         file['version'] = version
         items = _download_items(file, directory)
-        update_tasks.append(UpdateTask(file['path'], items))
+        is_latest_version = project_version == latest_proj_info["version"]
+        update_tasks.append(UpdateTask(file['path'], items, latest_version=is_latest_version))
 
     # make a single list of items to download
     total_size = 0
@@ -146,9 +153,9 @@ def download_project_async(mc, project_path, directory, project_version=None):
             total_size += item.size
 
     mp.log.info(f"will download {len(update_tasks)} files in {len(download_list)} chunks, total size {total_size}")
-    
+
     job = DownloadJob(project_path, total_size, version, update_tasks, download_list, directory, mp, project_info)
-    
+
     # start download
     job.executor = concurrent.futures.ThreadPoolExecutor(max_workers=4)
     job.futures = []
@@ -161,7 +168,7 @@ def download_project_async(mc, project_path, directory, project_version=None):
 
 def download_project_wait(job):
     """ blocks until all download tasks are finished """
-    
+
     concurrent.futures.wait(job.futures)
 
 
@@ -202,10 +209,10 @@ def download_project_finalize(job):
     job.mp.log.info("--- download finished")
 
     for task in job.update_tasks:
-        
+
         # right now only copy tasks...
         task.apply(job.directory, job.mp)
-    
+
     # final update of project metadata
     # TODO: why not exact copy of project info JSON ?
     job.mp.metadata = {
@@ -232,16 +239,17 @@ class UpdateTask:
     Entry for each file that will be updated.
     At the end of a successful download of new data, all the tasks are executed.
     """
-    
+
     # TODO: methods other than COPY
-    def __init__(self, file_path, download_queue_items, destination_file=None):
+    def __init__(self, file_path, download_queue_items, destination_file=None, latest_version=True):
         self.file_path = file_path
         self.destination_file = destination_file
         self.download_queue_items = download_queue_items
+        self.latest_version = latest_version
 
     def apply(self, directory, mp):
         """ assemble downloaded chunks into a single file """
-        
+
         if self.destination_file is None:
             basename = os.path.basename(self.file_path)
             file_dir = os.path.dirname(os.path.normpath(os.path.join(directory, self.file_path)))
@@ -251,8 +259,10 @@ class UpdateTask:
             dest_file_path = self.destination_file
         os.makedirs(file_dir, exist_ok=True)
 
+        # ignore check if we download not-latest version of gpkg file (possibly reconstructed on server on demand)
+        check_size = self.latest_version or not mp.is_versioned_file(self.file_path)
         # merge chunks together (and delete them afterwards)
-        file_to_merge = FileToMerge(dest_file_path, self.download_queue_items)
+        file_to_merge = FileToMerge(dest_file_path, self.download_queue_items, check_size)
         file_to_merge.merge()
 
         # Make a copy of the file to meta dir only if there is no user-specified path for the file.
@@ -263,7 +273,7 @@ class UpdateTask:
 
 class DownloadQueueItem:
     """ a piece of data from a project that should be downloaded - it can be either a chunk or it can be a diff """
-    
+
     def __init__(self, file_path, size, version, diff_only, part_index, download_file_path):
         self.file_path = file_path     # relative path to the file within project
         self.size = size               # size of the item in bytes
@@ -271,7 +281,7 @@ class DownloadQueueItem:
         self.diff_only = diff_only     # whether downloading diff or full version
         self.part_index = part_index    # index of the chunk
         self.download_file_path = download_file_path   # full path to a temporary file which will receive the content
-        
+
     def __repr__(self):
         return "<DownloadQueueItem path={} version={} diff_only={} part_index={} size={} dest={}>".format(
             self.file_path, self.version, self.diff_only, self.part_index, self.size, self.download_file_path)
@@ -299,7 +309,7 @@ class DownloadQueueItem:
 
 class PullJob:
     def __init__(self, project_path, pull_changes, total_size, version, files_to_merge, download_queue_items,
-                 temp_dir, mp, project_info, basefiles_to_patch):
+                 temp_dir, mp, project_info, basefiles_to_patch, user_name):
         self.project_path = project_path
         self.pull_changes = pull_changes    # dictionary with changes (dict[str, list[dict]] - keys: "added", "updated", ...)
         self.total_size = total_size      # size of data to download (in bytes)
@@ -312,6 +322,7 @@ class PullJob:
         self.is_cancelled = False
         self.project_info = project_info   # parsed JSON with project info returned from the server
         self.basefiles_to_patch = basefiles_to_patch  # list of tuples (relative path within project, list of diff files in temp dir to apply)
+        self.user_name = user_name
 
     def dump(self):
         print("--- JOB ---", self.total_size, "bytes")
@@ -333,6 +344,14 @@ def pull_project_async(mc, directory):
     """
 
     mp = MerginProject(directory)
+    if mp.has_unfinished_pull():
+        try:
+            mp.resolve_unfinished_pull(mc.username())
+        except ClientError as err:
+            mp.log.error("Error resolving unfinished pull: " + str(err))
+            mp.log.info("--- pull aborted")
+            raise
+
     project_path = mp.metadata["name"]
     local_version = mp.metadata["version"]
 
@@ -387,7 +406,7 @@ def pull_project_async(mc, directory):
     for file in fetch_files:
         diff_only = _pulling_file_with_diffs(file)
         items = _download_items(file, temp_dir, diff_only)
-        
+
         # figure out destination path for the file
         file_dir = os.path.dirname(os.path.normpath(os.path.join(temp_dir, file['path'])))
         basename = os.path.basename(file['diff']['path']) if diff_only else os.path.basename(file['path'])
@@ -429,7 +448,7 @@ def pull_project_async(mc, directory):
 
     mp.log.info(f"will download {len(download_list)} chunks, total size {total_size}")
 
-    job = PullJob(project_path, pull_changes, total_size, server_version, files_to_merge, download_list, temp_dir, mp, server_info, basefiles_to_patch)
+    job = PullJob(project_path, pull_changes, total_size, server_version, files_to_merge, download_list, temp_dir, mp, server_info, basefiles_to_patch, mc.username())
 
     # start download
     job.executor = concurrent.futures.ThreadPoolExecutor(max_workers=4)
@@ -437,13 +456,13 @@ def pull_project_async(mc, directory):
     for item in download_list:
         future = job.executor.submit(_do_download, item, mc, mp, project_path, job)
         job.futures.append(future)
-    
+
     return job
 
 
 def pull_project_wait(job):
     """ blocks until all download tasks are finished """
-    
+
     concurrent.futures.wait(job.futures)
 
 
@@ -483,9 +502,10 @@ class FileToMerge:
     to the temporary file containing its data. Calling merge() will create the destination file
     and remove the temporary files of the chunks
     """
-    def __init__(self, dest_file, downloaded_items):
+    def __init__(self, dest_file, downloaded_items, size_check=True):
         self.dest_file = dest_file   # full path to the destination file to be created
         self.downloaded_items = downloaded_items   # list of pieces of the destination file to be merged
+        self.size_check = size_check  # whether we want to do merged file size check
 
     def merge(self):
         with open(self.dest_file, 'wb') as final:
@@ -494,6 +514,8 @@ class FileToMerge:
                     shutil.copyfileobj(chunk, final)
                 os.remove(item.download_file_path)
 
+        if not self.size_check:
+            return
         expected_size = sum(item.size for item in self.downloaded_items)
         if os.path.getsize(self.dest_file) != expected_size:
             os.remove(self.dest_file)
@@ -503,13 +525,13 @@ class FileToMerge:
 def pull_project_finalize(job):
     """
     To be called when pull in the background is finished and we need to do the finalization (merge chunks etc.)
-    
+
     This should not be called from a worker thread (e.g. directly from a handler when download is complete)
 
     If any of the workers has thrown any exception, it will be re-raised (e.g. some network errors).
     That also means that the whole job has been aborted.
     """
-    
+
     job.executor.shutdown(wait=True)
 
     # make sure any exceptions from threads are not lost
@@ -552,14 +574,23 @@ def pull_project_finalize(job):
             os.remove(basefile)
             raise ClientError("Cannot patch basefile {}! Please try syncing again.".format(basefile))
 
-    conflicts = job.mp.apply_pull_changes(job.pull_changes, job.temp_dir)
+    try:
+        conflicts = job.mp.apply_pull_changes(job.pull_changes, job.temp_dir, job.user_name)
+    except Exception as e:
+        job.mp.log.error("Failed to apply pull changes: " + str(e))
+        job.mp.log.info("--- pull aborted")
+        raise ClientError("Failed to apply pull changes: " + str(e))
+
     job.mp.metadata = {
         'name': job.project_path,
         'version': job.version if job.version else "v0",  # for new projects server version is ""
         'files': job.project_info['files']
     }
 
-    job.mp.log.info("--- pull finished -- at version " + job.mp.metadata['version'])
+    if job.mp.has_unfinished_pull():
+        job.mp.log.info("--- failed to complete pull -- project left in the unfinished pull state")
+    else:
+        job.mp.log.info("--- pull finished -- at version " + job.mp.metadata['version'])
 
     shutil.rmtree(job.temp_dir)
     return conflicts
@@ -574,7 +605,11 @@ def download_file_async(mc, project_dir, file_path, output_file, version):
     project_path = mp.metadata["name"]
     ver_info = f"at version {version}" if version is not None else "at latest version"
     mp.log.info(f"Getting {file_path} {ver_info}")
-    project_info = mc.project_info(project_path, version=version)
+    latest_proj_info = mc.project_info(project_path)
+    if version:
+        project_info = mc.project_info(project_path, version=version)
+    else:
+        project_info = latest_proj_info
     mp.log.info(f"Got project info. version {project_info['version']}")
 
     # set temporary directory for download
@@ -587,7 +622,8 @@ def download_file_async(mc, project_dir, file_path, output_file, version):
         if file["path"] == file_path:
             file['version'] = version
             items = _download_items(file, temp_dir)
-            task = UpdateTask(file['path'], items, output_file)
+            is_latest_version = version == latest_proj_info["version"]
+            task = UpdateTask(file['path'], items, output_file, latest_version=is_latest_version)
             download_list.extend(task.download_queue_items)
             for item in task.download_queue_items:
                 total_size += item.size
@@ -636,7 +672,7 @@ def download_file_finalize(job):
         shutil.rmtree(temp_dir)
 
 
-def download_diffs_async(mc, project_directory, file_path, version_from, version_to):
+def download_diffs_async(mc, project_directory, file_path, versions):
     """
     Starts background download project file diffs for specified versions.
     Returns handle to the pending download.
@@ -645,8 +681,7 @@ def download_diffs_async(mc, project_directory, file_path, version_from, version
         mc (MerginClient): MerginClient instance.
         project_directory (str): local project directory.
         file_path (str): file path relative to Mergin project root.
-        version_from (str): starting project version tag for getting diff, for example 'v3'.
-        version_to (str): ending project version tag for getting diff.
+        versions (list): list of versions to download diffs for, e.g. ['v1', 'v2'].
 
     Returns:
         PullJob/None: a handle for the pending download.
@@ -654,11 +689,8 @@ def download_diffs_async(mc, project_directory, file_path, version_from, version
     mp = MerginProject(project_directory)
     project_path = mp.metadata["name"]
     file_history = mc.project_file_history_info(project_path, file_path)
-    versions_to_fetch = get_versions_with_file_changes(
-        mc, project_path, file_path, version_from=version_from, version_to=version_to, file_history=file_history
-    )
     mp.log.info(f"--- version: {mc.user_agent_info()}")
-    mp.log.info(f"--- start download diffs for {file_path} of {project_path}, versions: {[v for v in versions_to_fetch]}")
+    mp.log.info(f"--- start download diffs for {file_path} of {project_path}, versions: {[v for v in versions]}")
 
     try:
         server_info = mc.project_info(project_path)
@@ -669,11 +701,12 @@ def download_diffs_async(mc, project_directory, file_path, version_from, version
         mp.log.info("--- downloading diffs aborted")
         raise
 
-    temp_dir = tempfile.mkdtemp(prefix="mergin-py-client-")
     fetch_files = []
 
-    for version in versions_to_fetch[1:]:
+    for version in versions:
         version_data = file_history["history"][version]
+        if "diff" not in version_data:
+            continue  # skip if there is no diff in history
         diff_data = copy.deepcopy(version_data)
         diff_data['version'] = version
         diff_data['diff'] = version_data['diff']
@@ -683,8 +716,10 @@ def download_diffs_async(mc, project_directory, file_path, version_from, version
     download_list = []  # list of all items to be downloaded
     total_size = 0
     for file in fetch_files:
-        items = _download_items(file, temp_dir, diff_only=True)
-        dest_file_path = os.path.normpath(os.path.join(temp_dir, os.path.basename(file['diff']['path'])))
+        items = _download_items(file, mp.cache_dir, diff_only=True)
+        dest_file_path = mp.fpath_cache(file['diff']['path'], version=file["version"])
+        if os.path.exists(dest_file_path):
+            continue
         files_to_merge.append(FileToMerge(dest_file_path, items))
         download_list.extend(items)
         for item in items:
@@ -692,8 +727,8 @@ def download_diffs_async(mc, project_directory, file_path, version_from, version
 
     mp.log.info(f"will download {len(download_list)} chunks, total size {total_size}")
 
-    job = PullJob(project_path, None, total_size, None, files_to_merge, download_list, temp_dir, mp,
-                  server_info, {})
+    job = PullJob(project_path, None, total_size, None, files_to_merge, download_list, mp.cache_dir, mp,
+                  server_info, {}, mc.username())
 
     # start download
     job.executor = concurrent.futures.ThreadPoolExecutor(max_workers=4)
@@ -705,8 +740,12 @@ def download_diffs_async(mc, project_directory, file_path, version_from, version
     return job
 
 
-def download_diffs_finalize(job, output_diff):
-    """ To be called after download_diffs_async """
+def download_diffs_finalize(job):
+    """ To be called after download_diffs_async
+
+    Returns:
+        diffs: list of downloaded diffs (their actual locations on disk)
+    """
 
     job.executor.shutdown(wait=True)
 
@@ -718,35 +757,17 @@ def download_diffs_finalize(job, output_diff):
             raise future.exception()
 
     job.mp.log.info("finalizing diffs pull")
+    diffs = []
 
     # merge downloaded chunks
     try:
         for file_to_merge in job.files_to_merge:
             file_to_merge.merge()
+            diffs.append(file_to_merge.dest_file)
     except ClientError as err:
         job.mp.log.error("Error merging chunks of downloaded file: " + str(err))
         job.mp.log.info("--- diffs pull aborted")
         raise
 
     job.mp.log.info("--- diffs pull finished")
-
-    # Collect and finally concatenate diffs, if needed
-    diffs = []
-    for file_to_merge in job.files_to_merge:
-        diffs.append(file_to_merge.dest_file)
-
-    output_dir = os.path.dirname(output_diff)
-    temp_dir = None
-    if len(diffs) >= 1:
-        os.makedirs(output_dir, exist_ok=True)
-        temp_dir = os.path.dirname(diffs[0])
-        if len(diffs) > 1:
-            job.mp.geodiff.concat_changes(diffs, output_diff)
-        elif len(diffs) == 1:
-            shutil.copy(diffs[0], output_diff)
-    for diff in diffs:
-        os.remove(diff)
-
-    # remove the diffs download temporary directory
-    if temp_dir is not None:
-        shutil.rmtree(temp_dir)
+    return diffs
