@@ -9,10 +9,18 @@ import pytest
 import pytz
 import sqlite3
 
+from .. import InvalidProject
 from ..client import MerginClient, ClientError, MerginProject, LoginError, decode_token_data, TokenError
 from ..client_push import push_project_async, push_project_cancel
-from ..utils import generate_checksum, get_versions_with_file_changes
+from ..utils import (
+    generate_checksum,
+    get_versions_with_file_changes,
+    unique_path_name,
+    conflicted_copy_file_name,
+    edit_conflict_file_name
+)
 from ..merginproject import pygeodiff
+from ..report import create_report
 
 
 SERVER_URL = os.environ.get('TEST_MERGIN_URL')
@@ -133,7 +141,7 @@ def test_create_remote_project_from_local(mc):
         assert f in downloads['dir']
         if mp.is_versioned_file(f):
             assert f in downloads['meta']
-    
+
     # unable to download to the same directory
     with pytest.raises(Exception, match='Project directory already exists'):
         mc.download_project(project, download_dir)
@@ -189,7 +197,7 @@ def test_push_pull_changes(mc):
     assert len(project_info['files']) == len(mp.inspect_files())
     project_versions = mc.project_versions(project)
     assert len(project_versions) == 2
-    f_change = next((f for f in project_versions[0]['changes']['updated'] if f['path'] == f_updated), None)
+    f_change = next((f for f in project_versions[-1]['changes']['updated'] if f['path'] == f_updated), None)
     assert 'origin_checksum' not in f_change  # internal client info
 
     # test parallel changes
@@ -214,8 +222,8 @@ def test_push_pull_changes(mc):
     assert not os.path.exists(os.path.join(project_dir_2, f_removed))
     assert not os.path.exists(os.path.join(project_dir_2, f_renamed))
     assert os.path.exists(os.path.join(project_dir_2, 'renamed.txt'))
-    assert os.path.exists(os.path.join(project_dir_2, f_updated+'_conflict_copy'))
-    assert generate_checksum(os.path.join(project_dir_2, f_updated+'_conflict_copy')) == f_conflict_checksum
+    assert os.path.exists(os.path.join(project_dir_2, conflicted_copy_file_name(f_updated, API_USER, 1)))
+    assert generate_checksum(os.path.join(project_dir_2, conflicted_copy_file_name(f_updated, API_USER, 1))) == f_conflict_checksum
     assert generate_checksum(os.path.join(project_dir_2, f_updated)) == f_remote_checksum
 
 
@@ -588,7 +596,7 @@ def test_available_storage_validation(mc):
     except ClientError as e:
         # Expecting "Storage limit has been reached" error msg.
         assert str(e).startswith("Storage limit has been reached")
-        got_right_err = True    
+        got_right_err = True
     assert got_right_err
 
     # Expecting empty project
@@ -811,7 +819,7 @@ def test_server_compatibility(mc):
     assert mc.is_server_compatible()
 
 
-def create_versioned_project(mc, project_name, project_dir, updated_file, remove=True):
+def create_versioned_project(mc, project_name, project_dir, updated_file, remove=True, overwrite=False):
     project = API_USER + '/' + project_name
     cleanup(mc, project, [project_dir])
 
@@ -831,6 +839,12 @@ def create_versioned_project(mc, project_name, project_dir, updated_file, remove
         os.remove(os.path.join(project_dir, updated_file))
         mc.push_project(project_dir)
 
+    # create version with forced overwrite (broken history)
+    if overwrite:
+        shutil.move(mp.fpath(updated_file), mp.fpath_meta(updated_file))
+        shutil.copy(os.path.join(CHANGED_SCHEMA_DIR, 'modified_schema.gpkg'), mp.fpath(updated_file))
+        shutil.copy(os.path.join(CHANGED_SCHEMA_DIR, 'modified_schema.gpkg-wal'), mp.fpath(updated_file + '-wal'))
+        mc.push_project(project_dir)
     return mp
 
 
@@ -941,6 +955,43 @@ def test_download_diffs(mc):
     assert "Available versions: [1, 2, 3, 4]" in str(e.value)
 
 
+def test_modify_project_permissions(mc):
+    test_project = 'test_project'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)
+    download_dir = os.path.join(TMP_DIR, 'download', test_project)
+
+    cleanup(mc, project, [project_dir, download_dir])
+    # prepare local project
+    shutil.copytree(TEST_DATA_DIR, project_dir)
+
+    # create remote project
+    mc.create_project_and_push(test_project, directory=project_dir)
+
+    # check basic metadata about created project
+    project_info = mc.project_info(project)
+    assert project_info['version'] == 'v1'
+    assert project_info['name'] == test_project
+    assert project_info['namespace'] == API_USER
+
+    permissions = mc.project_user_permissions(project)
+    assert permissions["owners"] == [API_USER]
+    assert permissions["writers"] == [API_USER]
+    assert permissions["readers"] == [API_USER]
+
+    mc.add_user_permissions_to_project(project, [API_USER2], "writer")
+    permissions = mc.project_user_permissions(project)
+    assert set(permissions["owners"]) == {API_USER}
+    assert set(permissions["writers"]) == {API_USER, API_USER2}
+    assert set(permissions["readers"]) == {API_USER, API_USER2}
+
+    mc.remove_user_permissions_from_project(project, [API_USER2])
+    permissions = mc.project_user_permissions(project)
+    assert permissions["owners"] == [API_USER]
+    assert permissions["writers"] == [API_USER]
+    assert permissions["readers"] == [API_USER]
+
+
 def _use_wal(db_file):
     """ Ensures that sqlite database is using WAL journal mode """
     con = sqlite3.connect(db_file)
@@ -1098,7 +1149,7 @@ def test_rebase_local_schema_change(mc, extra_connection):
     project_dir_2 = os.path.join(TMP_DIR, test_project+'_2')  # concurrent project dir
     test_gpkg = os.path.join(project_dir, 'test.gpkg')
     test_gpkg_basefile = os.path.join(project_dir, '.mergin', 'test.gpkg')
-    test_gpkg_conflict = test_gpkg + '_conflict_copy'
+    test_gpkg_conflict = conflicted_copy_file_name(test_gpkg, API_USER, 1)
     cleanup(mc, project, [project_dir, project_dir_2])
 
     os.makedirs(project_dir)
@@ -1164,7 +1215,7 @@ def test_rebase_remote_schema_change(mc, extra_connection):
     test_gpkg = os.path.join(project_dir, 'test.gpkg')
     test_gpkg_2 = os.path.join(project_dir_2, 'test.gpkg')
     test_gpkg_basefile = os.path.join(project_dir, '.mergin', 'test.gpkg')
-    test_gpkg_conflict = test_gpkg + '_conflict_copy'
+    test_gpkg_conflict = conflicted_copy_file_name(test_gpkg, API_USER, 1)
     cleanup(mc, project, [project_dir, project_dir_2])
 
     os.makedirs(project_dir)
@@ -1229,7 +1280,7 @@ def test_rebase_success(mc, extra_connection):
     test_gpkg = os.path.join(project_dir, 'test.gpkg')
     test_gpkg_2 = os.path.join(project_dir_2, 'test.gpkg')
     test_gpkg_basefile = os.path.join(project_dir, '.mergin', 'test.gpkg')
-    test_gpkg_conflict = test_gpkg + '_conflict_copy'
+    test_gpkg_conflict = conflicted_copy_file_name(test_gpkg, API_USER, 1)
     cleanup(mc, project, [project_dir, project_dir_2])
 
     os.makedirs(project_dir)
@@ -1267,3 +1318,388 @@ def test_rebase_success(mc, extra_connection):
 
     assert _get_table_row_count(test_gpkg, 'simple') == 5
     assert _get_table_row_count(test_gpkg_basefile, 'simple') == 4
+
+
+def test_conflict_file_names():
+    """
+    Test generation of file names for conflicts files.
+    """
+
+    data = [
+            ('/home/test/geo.gpkg', 'jack', 10, '/home/test/geo (conflicted copy, jack v10).gpkg'),
+            ('/home/test/g.pkg', 'j', 0, '/home/test/g (conflicted copy, j v0).pkg'),
+            ('home/test/geo.gpkg', 'jack', 10, 'home/test/geo (conflicted copy, jack v10).gpkg'),
+            ('geo.gpkg', 'jack', 10, 'geo (conflicted copy, jack v10).gpkg'),
+            ('/home/../geo.gpkg', 'jack', 10, '/geo (conflicted copy, jack v10).gpkg'),
+            ('/home/./geo.gpkg', 'jack', 10, '/home/geo (conflicted copy, jack v10).gpkg'),
+            ('/home/test/geo.gpkg', '', 10, '/home/test/geo (conflicted copy,  v10).gpkg'),
+            ('/home/test/geo.gpkg', 'jack', -1, '/home/test/geo (conflicted copy, jack v-1).gpkg'),
+            ('/home/test/geo.tar.gz', 'jack', 100, '/home/test/geo (conflicted copy, jack v100).tar.gz'),
+            ('', 'jack', 1, '' )
+           ]
+
+    for i in data:
+        file_name = conflicted_copy_file_name(i[0], i[1], i[2])
+        assert file_name == i[3]
+
+
+    data = [
+            ('/home/test/geo.json', 'jack', 10, '/home/test/geo (edit conflict, jack v10).json'),
+            ('/home/test/g.jsn', 'j', 0, '/home/test/g (edit conflict, j v0).json'),
+            ('home/test/geo.json', 'jack', 10, 'home/test/geo (edit conflict, jack v10).json'),
+            ('geo.json', 'jack', 10, 'geo (edit conflict, jack v10).json'),
+            ('/home/../geo.json', 'jack', 10, '/geo (edit conflict, jack v10).json'),
+            ('/home/./geo.json', 'jack', 10, '/home/geo (edit conflict, jack v10).json'),
+            ('/home/test/geo.json', '', 10, '/home/test/geo (edit conflict,  v10).json'),
+            ('/home/test/geo.json', 'jack', -1, '/home/test/geo (edit conflict, jack v-1).json'),
+            ('/home/test/geo.gpkg', 'jack', 10, '/home/test/geo (edit conflict, jack v10).json'),
+            ('/home/test/geo.tar.gz', 'jack', 100, '/home/test/geo (edit conflict, jack v100).json'),
+            ('', 'jack', 1, '')
+           ]
+
+    for i in data:
+        file_name = edit_conflict_file_name(i[0], i[1], i[2])
+        assert file_name == i[3]
+
+
+def test_unique_path_names():
+    """
+    Test generation of unique file names.
+    """
+    project_dir = os.path.join(TMP_DIR, 'unique_file_names')
+
+    remove_folders([project_dir])
+
+    os.makedirs(project_dir)
+    assert os.path.exists(project_dir)
+    assert not any(os.scandir(project_dir))
+
+    # Create test directory structure:
+    # - folderA
+    #   |- fileA.txt
+    #   |- fileA (1).txt
+    #   |- fileB.txt
+    #   |- folderAB
+    #   |- folderAB (1)
+    # - file.txt
+    # - another.txt
+    # - another (1).txt
+    # - another (2).txt
+    # - arch.tar.gz
+    data = {'folderA': {'files': ['fileA.txt', 'fileA (1).txt', 'fileB.txt'], 'folderAB': {}, 'folderAB (1)': {}}, 'files': ['file.txt', 'another.txt', 'another (1).txt', 'another (2).txt', 'arch.tar.gz']}
+    create_directory(project_dir, data)
+
+    data = [
+            ('file.txt', 'file (1).txt'),
+            ('another.txt', 'another (3).txt'),
+            ('folderA', 'folderA (1)'),
+            ('non.txt', 'non.txt'),
+            ('data.gpkg', 'data.gpkg'),
+            ('arch.tar.gz', 'arch (1).tar.gz'),
+            ('folderA/folder', 'folderA/folder'),
+            ('folderA/fileA.txt', 'folderA/fileA (2).txt'),
+            ('folderA/fileB.txt', 'folderA/fileB (1).txt'),
+            ('folderA/fileC.txt', 'folderA/fileC.txt'),
+            ('folderA/folderAB', 'folderA/folderAB (2)'),
+           ]
+    for i in data:
+        file_name = unique_path_name(os.path.join(project_dir, i[0]))
+        assert file_name == os.path.join(project_dir, i[1])
+
+def create_directory(root, data):
+    for k, v in data.items():
+        if isinstance(v, dict):
+            dir_name = os.path.join(root, k)
+            os.makedirs(dir_name, exist_ok=True)
+            for kk in v.keys():
+                create_directory(dir_name, v)
+        elif isinstance(v, list):
+            for file_name in v:
+                open(os.path.join(root, file_name), 'w').close()
+
+def test_unfinished_pull(mc):
+    """
+    Checks whether a pull with failed rebase (due to remote DB schema change)
+    and failed copy of the database file is handled correctly, i.e. an
+    unfinished_pull directory is created with the content of the server changes.
+    """
+    test_project = 'test_unfinished_pull'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)  # primary project dir
+    project_dir_2 = os.path.join(TMP_DIR, test_project+'_2')  # concurrent project dir
+    unfinished_pull_dir = os.path.join(TMP_DIR, test_project, '.mergin', 'unfinished_pull')  # unfinished_pull dir for the primary project
+    test_gpkg = os.path.join(project_dir, 'test.gpkg')
+    test_gpkg_2 = os.path.join(project_dir_2, 'test.gpkg')
+    test_gpkg_basefile = os.path.join(project_dir, '.mergin', 'test.gpkg')
+    test_gpkg_conflict = conflicted_copy_file_name(test_gpkg, API_USER, 2)
+    test_gpkg_unfinished_pull = os.path.join(project_dir, '.mergin', 'unfinished_pull', 'test.gpkg')
+    cleanup(mc, project, [project_dir, project_dir_2])
+
+    os.makedirs(project_dir)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'base.gpkg'), test_gpkg)
+    _use_wal(test_gpkg)  # make sure we use WAL, that's the more common and more difficult scenario
+    mc.create_project_and_push(test_project, project_dir)
+
+    # Download project to the concurrent dir + change DB schema + push a new version
+    mc.download_project(project, project_dir_2)
+    _create_test_table(test_gpkg_2)
+    mc.push_project(project_dir_2)
+
+    # do changes in the local DB (added a row)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'inserted_1_A.gpkg'), test_gpkg)
+    _use_wal(test_gpkg)  # make sure we use WAL
+
+    pull_changes, push_changes, _ = mc.project_status(project_dir)
+    assert _is_file_updated('test.gpkg', pull_changes)
+    assert _is_file_updated('test.gpkg', push_changes)
+
+    assert not os.path.exists(test_gpkg_conflict)
+    assert not mc.has_unfinished_pull(project_dir)
+
+    # lock base file to emulate situation when we can't overwrite it, because
+    # it is used by another process
+    subprocess.run(['sudo', 'chattr', '+i', test_gpkg])
+
+    mc.pull_project(project_dir)
+
+    assert not os.path.exists(test_gpkg_conflict)
+    assert mc.has_unfinished_pull(project_dir)
+
+    # check the results after pull:
+    # - unfinished pull file should contain the new table
+    # - local file + basefile should not contain the new table
+    _check_test_table(test_gpkg_unfinished_pull)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg_basefile)
+
+    # check that the local file contain the new row, while basefile and server version don't
+    assert _get_table_row_count(test_gpkg, 'simple') == 4
+    assert _get_table_row_count(test_gpkg_basefile, 'simple') == 3
+    assert _get_table_row_count(test_gpkg_unfinished_pull, 'simple') == 3
+
+    # unlock base file, so we can apply changes from the unfinished pull
+    subprocess.run(['sudo', 'chattr', '-i', test_gpkg])
+
+    mc.resolve_unfinished_pull(project_dir)
+
+    assert os.path.exists(test_gpkg_conflict)
+    assert not mc.has_unfinished_pull(project_dir)
+
+    # check the results after resolving unfinished pull:
+    # - conflict copy should not contain the new table
+    # - local file + basefile should contain the new table
+    _check_test_table(test_gpkg)
+    _check_test_table(test_gpkg_basefile)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg_conflict)
+
+    # check that the local file + basefile don't contain the new row, and the conflict copy does
+    assert _get_table_row_count(test_gpkg, 'simple') == 3
+    assert _get_table_row_count(test_gpkg_basefile, 'simple') == 3
+    assert _get_table_row_count(test_gpkg_conflict, 'simple') == 4
+
+def test_unfinished_pull_push(mc):
+    """
+    Checks client behaviour when performing push and pull of the project
+    in the unfinished pull state.
+    """
+    test_project = 'test_unfinished_pull_push'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)  # primary project dir
+    project_dir_2 = os.path.join(TMP_DIR, test_project+'_2')  # concurrent project dir
+    unfinished_pull_dir = os.path.join(TMP_DIR, test_project, '.mergin', 'unfinished_pull')  # unfinished_pull dir for the primary project
+    test_gpkg = os.path.join(project_dir, 'test.gpkg')
+    test_gpkg_2 = os.path.join(project_dir_2, 'test.gpkg')
+    test_gpkg_basefile = os.path.join(project_dir, '.mergin', 'test.gpkg')
+    test_gpkg_conflict = conflicted_copy_file_name(test_gpkg, API_USER, 2)
+    test_gpkg_unfinished_pull = os.path.join(project_dir, '.mergin', 'unfinished_pull', 'test.gpkg')
+    cleanup(mc, project, [project_dir, project_dir_2])
+
+    os.makedirs(project_dir)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'base.gpkg'), test_gpkg)
+    _use_wal(test_gpkg)  # make sure we use WAL, that's the more common and more difficult scenario
+    mc.create_project_and_push(test_project, project_dir)
+
+    # Download project to the concurrent dir + change DB schema + push a new version
+    mc.download_project(project, project_dir_2)
+    _create_test_table(test_gpkg_2)
+    mc.push_project(project_dir_2)
+
+    # do changes in the local DB (added a row)
+    shutil.copy(os.path.join(TEST_DATA_DIR, 'inserted_1_A.gpkg'), test_gpkg)
+    _use_wal(test_gpkg)  # make sure we use WAL
+
+    pull_changes, push_changes, _ = mc.project_status(project_dir)
+    assert _is_file_updated('test.gpkg', pull_changes)
+    assert _is_file_updated('test.gpkg', push_changes)
+
+    assert not os.path.exists(test_gpkg_conflict)
+    assert not mc.has_unfinished_pull(project_dir)
+
+    # lock base file to emulate situation when we can't overwrite it, because
+    # it is used by another process
+    subprocess.run(['sudo', 'chattr', '+i', test_gpkg])
+
+    mc.pull_project(project_dir)
+
+    assert not os.path.exists(test_gpkg_conflict)
+    assert mc.has_unfinished_pull(project_dir)
+
+    # check the results after pull:
+    # - unfinished pull file should contain the new table
+    # - local file + basefile should not contain the new table
+    _check_test_table(test_gpkg_unfinished_pull)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg_basefile)
+
+    # check that the local file contain the new row, while basefile and server version don't
+    assert _get_table_row_count(test_gpkg, 'simple') == 4
+    assert _get_table_row_count(test_gpkg_basefile, 'simple') == 3
+    assert _get_table_row_count(test_gpkg_unfinished_pull, 'simple') == 3
+
+    # attempt to push project in the unfinished pull state should
+    # fail with ClientError
+    with pytest.raises(ClientError):
+        mc.push_project(project_dir)
+
+    # attempt to pull project in the unfinished pull state should
+    # fail with ClientError if unfinished pull can not be resolved
+    with pytest.raises(ClientError):
+        mc.pull_project(project_dir)
+
+    # unlock base file, so we can apply changes from the unfinished pull
+    subprocess.run(['sudo', 'chattr', '-i', test_gpkg])
+
+    # perform pull. This should resolve unfinished pull first and then
+    # collect data from the server
+    mc.pull_project(project_dir)
+
+    assert os.path.exists(test_gpkg_conflict)
+    assert not mc.has_unfinished_pull(project_dir)
+
+    # check the results after resolving unfinished pull:
+    # - conflict copy should not contain the new table
+    # - local file + basefile should contain the new table
+    _check_test_table(test_gpkg)
+    _check_test_table(test_gpkg_basefile)
+    with pytest.raises(sqlite3.OperationalError):
+        _check_test_table(test_gpkg_conflict)
+
+    # check that the local file + basefile don't contain the new row, and the conflict copy does
+    assert _get_table_row_count(test_gpkg, 'simple') == 3
+    assert _get_table_row_count(test_gpkg_basefile, 'simple') == 3
+    assert _get_table_row_count(test_gpkg_conflict, 'simple') == 4
+
+def test_project_versions_list(mc):
+    """Test getting project versions in various ranges """
+    test_project = 'test_project_versions'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)
+    create_versioned_project(mc, test_project, project_dir, "base.gpkg")
+    project_info = mc.project_info(project)
+    assert project_info["version"] == "v5"
+
+    # get all versions
+    versions = mc.project_versions(project)
+    assert len(versions) == 5
+    assert versions[0]["name"] == "v1"
+    assert versions[-1]["name"] == "v5"
+
+    # get first 3 versions
+    versions = mc.project_versions(project, to="v3")
+    assert len(versions) == 3
+    assert versions[-1]["name"] == "v3"
+
+    # get last 2 versions
+    versions = mc.project_versions(project, since="v4")
+    assert len(versions) == 2
+    assert versions[0]["name"] == "v4"
+
+    # get range v2-v4
+    versions = mc.project_versions(project, since="v2", to="v4")
+    assert len(versions) == 3
+    assert versions[0]["name"] == "v2"
+    assert versions[-1]["name"] == "v4"
+
+def test_report(mc):
+    test_project = 'test_download_diffs'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)
+    f_updated = "base.gpkg"
+    mp = create_versioned_project(mc, test_project, project_dir, f_updated, remove=False, overwrite=True)
+
+    # create report for between versions 2 and 4
+    directory = mp.dir
+    since = "v2"
+    to = "v4"
+    proj_name = project.replace(os.path.sep, "-")
+    report_file = os.path.join(TMP_DIR, "report", f'{proj_name}-{since}-{to}.csv')
+    if os.path.exists(report_file):
+        os.remove(report_file)
+    warnings = create_report(mc, directory, since, to, report_file)
+    assert os.path.exists(report_file)
+    assert not warnings
+    # assert headers and content in report file
+    with open(report_file, "r") as rf:
+        content = rf.read()
+        headers = ",".join(["file", "table", "author", "date", "time", "version", "operation", "length", "area", "count"])
+        assert headers in content
+        assert "base.gpkg,simple,test_plugin" in content
+        assert "v3,update,,,2" in content
+        # files not edited are not in reports
+        assert "inserted_1_A.gpkg" not in content
+
+    # do report for v1 with added files and v5 with overwritten file
+    warnings = create_report(mc, directory, "v1", "v5", report_file)
+    assert warnings
+
+    # rm local mergin project and try again
+    shutil.rmtree(directory)
+    with pytest.raises(InvalidProject):
+        create_report(mc, directory, since, to, report_file)
+
+def test_project_versions_list(mc, mc2):
+    """
+    Test retrieving user permissions
+    """
+    test_project = 'test_permissions'
+    test_project_fullname = API_USER2 + '/' + test_project
+
+    # cleanups
+    project_dir = os.path.join(TMP_DIR, test_project, API_USER)
+    cleanup(mc, test_project_fullname, [project_dir])
+    cleanup(mc2, test_project_fullname, [project_dir])
+
+    # create new (empty) project on server
+    mc2.create_project(test_project)
+
+    # Add reader access to another client
+    project_info = get_project_info(mc2, API_USER2, test_project)
+    access = project_info['access']
+    access['readersnames'].append(API_USER)
+    mc2.set_project_access(test_project_fullname, access)
+
+    # reader should not have write access
+    assert not mc.has_writing_permissions(test_project_fullname)
+
+    # Add writer access to another client
+    project_info = get_project_info(mc2, API_USER2, test_project)
+    access = project_info['access']
+    access['writersnames'].append(API_USER)
+    mc2.set_project_access(test_project_fullname, access)
+
+    # now user shold have write access
+    assert mc.has_writing_permissions(test_project_fullname)
+
+    # test organization permissions
+    test_project_fullname = 'testorg' + '/' + 'test_org_permissions'
+
+    # owner should have write access
+    assert mc.has_writing_permissions(test_project_fullname)
+
+    # writer should have write access
+    assert mc2.has_writing_permissions(test_project_fullname)
